from copy import deepcopy

import requests


class Replacer(object):
    """ change or replace data from string, list or dict
    """

    def __init__(self, *rules, data=None, replace_key=False, mode='replace', dtype=None):
        self.rules = rules if data else rules[:-1]
        self._data = deepcopy(data) if data else rules[-1]
        self.replace_key = replace_key
        self.dtype = dtype
        self.search_result = {}
        self.mode = mode
        if self.mode in ['search', 'delete']:
            self._arg_keys = self.arg_keys()
        elif self.mode == 'update':
            self._update_dict = self.update_dict()

    @property
    def data(self):
        self._data = self.process(data=self._data)
        return self._data

    # --------------------------------------------------------------------------------- core function
    def process(self, key=None, data=None):
        """ 遍历整个数据 for search mode and delete mode

        :param data: dict or list
        :return: dict or list
        """
        if self.mode == 'delete':
            for k in self._arg_keys:
                try:
                    del data[k]
                except:
                    continue
        elif self.mode == 'search':
            if key and key in self._arg_keys:
                self.update_search_result(key, data)

        if isinstance(data, dict):
            # process dict data here

            for k, value in data.items():
                if not isinstance(value, (list, dict)):
                    data[k] = self.apply_rule(key=k, value=value)
                else:
                    if self.mode == 'update' and k in self._update_dict.keys():
                        data[k] = self._update_dict.get(k)
                    else:
                        data[k] = self.process(k, value)

        elif isinstance(data, list):
            # process list data here
            _data = []
            for d in data:

                if not isinstance(d, (list, dict)):
                    _data.append(self.apply_rule(value=d))
                else:
                    _data.append(self.process(data=d))

            return _data
        else:
            raise Exception('Data Type Error ... data type is unsupported')

        return data

    def apply_rule(self, key=None, value=None):
        """ 处理数据 process other type data here
        for replace, update, search mode

        :param key:str, int, float or others
        :param value: all types
        :return:value
        """

        if self.mode == 'replace':
            # replace police
            if isinstance(value, (int, float)):
                return value

            for rule in self.rules:
                if isinstance(rule, (int, float)): continue

                if isinstance(rule, (str, list)):
                    for r in list(rule):
                        if isinstance(r, (int, float)): continue
                        value = value.replace(r, '')
                elif isinstance(rule, dict):
                    for k, v in rule.items():
                        value = value.replace(k, v)

        elif self.mode == 'update':
            if key and key in self._update_dict.keys():
                value = self._update_dict.get(key)

        elif self.mode == 'search' and key and key in self._arg_keys:
            self.update_search_result(key, value)

        return value

    def arg_keys(self):
        keys = []
        for rule in self.rules:
            if isinstance(rule, str):
                keys.append(rule)
            elif isinstance(rule, list):
                keys.extend(rule)
            elif isinstance(rule, dict):
                keys.extend(rule.keys())
        return keys

    def update_search_result(self, key, value):
        if key in self.search_result.keys():
            if not isinstance(self.search_result.get(key), list):
                self.search_result[key] = [self.search_result.pop(key), value]
            elif not isinstance(value, list):
                self.search_result[key].append(value)
            else:
                self.search_result[key].extend(value)
        else:
            self.search_result[key] = value

    def update_dict(self):
        up_dict = {}
        if len(self.rules) == 1 and isinstance(self.rules[0], dict):
            up_dict = self.rules[0]

        elif len(self.rules) == 2 and isinstance(self.rules[0], str) and isinstance(self.rules[2], str):
            up_dict = dict([self.rules])

        elif len(self.rules) == 2 and isinstance(self.rules[0], list) and isinstance(self.rules[1], list):
            up_dict = dict(list(zip(self.rules)))

        elif len(self.rules) == 2 and isinstance(self.rules[0], list) and isinstance(self.rules[1], (str, int, float)):
            up_dict = dict((_, self.rules[1]) for _ in self.rules[0])

        elif len(self.rules) == 2 and isinstance(self.rules[0], dict) and isinstance(self.rules[0], dict):
            for k, v in self.rules[0].items():
                value = self.rules[1].get(v)
                up_dict[k] = value

        # list,list,dict
        elif len(self.rules) == 3 and isinstance(self.rules[0], list) and isinstance(self.rules[2], dict):
            for k, v in zip(self.rules[0], self.rules[1]):
                value = self.rules[2].get(v)
                up_dict[k] = value
        else:
            raise Exception('Args Format Error ... unsupported args format')

        return up_dict


class Printer(object):
    def __init__(self, data=None):
        self.data = data if data else ''
        self.data_groups = []
        self.information_extractor()
        self.printer()

    def printer(self):
        for dg in self.data_groups:
            lines = self.parse_data_group(dg)
            for line in lines:
                print(line)

    def information_extractor(self):
        if isinstance(self.data, requests.models.Response):
            data_dict = {
                'url': self.data.url,
                'body': self.data.request.body,
                'headers': self.data.request.headers,
                'cookies': self.data.request._cookies,
            }
            self.add_to_data_group('request', data_dict)

            data_dict = {
                'headers': self.data.headers,
                'cookies': self.data.cookies,
            }
            self.add_to_data_group('response', data_dict)

    def add_to_data_group(self, name, data_dict):
        group = DataGroup(name)
        for key, value in data_dict.items():
            if not value: value = ''
            if isinstance(value, str):
                group.add_info(key, value)
            else:
                group.add_data(key, value)
        self.data_groups.append(group)

    @staticmethod
    def parse_data_group(data_group):
        bar_length = data_group.bar_length
        head = f'{data_group.name} {"+" * (bar_length + len(data_group.name))}'
        info = data_group.info
        data_pool = data_group.data_pool
        lines = [head]

        # add info to lines
        for info_k, info_v in info.items():
            fmt = '{:<%d} | {:<%d}' % tuple(data_group.max_info_length)
            lines.append(fmt.format(info_k, info_v))

        lines.append('')
        # add data to lines
        for key, value in data_pool.items():
            lines.append(f'{key} {"-" * bar_length}')
            for d_key, d_value in value.items():
                fmt = '{:<%d} | {:<%d}' % tuple(data_group.max_data_length)
                lines.append(fmt.format(str(d_key), str(d_value)))

            lines.append('')
        return lines


class DataGroup(object):
    def __init__(self, name):
        self.name = name.upper()
        self.data_pool = {}
        self.info = {}
        self.max_info_length = [0, 0]
        self.max_data_length = [0, 0]
        self.bar_length = 0

    def add_data(self, title, data):
        if not data: return
        max_key_length = max([len(str(_)) for _ in data.keys()])
        max_value_length = max([len(str(_)) for _ in data.values()])

        self.max_data_length[0] = max_key_length if max_key_length > self.max_data_length[0] else self.max_data_length[
            0]
        self.max_data_length[1] = max_value_length if max_value_length > self.max_data_length[1] else \
            self.max_data_length[1]
        self.update_bar()

        self.data_pool[title] = data

    def add_info(self, title, info):
        self.max_info_length[0] = len(title) if len(title) > self.max_info_length[0] else self.max_info_length[0]
        self.max_info_length[1] = len(info) if len(info) > self.max_info_length[1] else self.max_info_length[1]
        self.info[title] = info
        self.update_bar()

    def update_bar(self):
        info_length = self.max_info_length[0] + self.max_info_length[1]
        data_length = self.max_data_length[0] + self.max_data_length[1]
        self.bar_length = info_length if info_length > data_length else data_length


class DictFactory(object):

    def del_dict_depth(self, data_dict):
        data = {}
        list_data = {}
        for key, value in data_dict.items():
            value = str(value)
            if isinstance(value, dict):
                data = {**data, **self.del_dict_depth(value)}
            elif isinstance(value, list):
                if not value:
                    continue
                if isinstance(value[0], dict):
                    list_data[key] = self.list_to_dict(value)
                else:
                    value = '|'.join(value)
                data = {**data, key: value}
            else:
                data = {**data, key: value}

        return {**data, **list_data}

    @staticmethod
    def list_to_dict(data_list):
        result = {}
        for data in data_list:
            for key, value in data.items():
                if key in [_ for _ in result.keys()]:
                    value_ = result.get(key)
                    result[key] = f'{value_}|{value}'
                else:
                    result = {**result, key: value}

        return result

    @staticmethod
    def find(target, dict_data, find_key=False):
        queue = [dict_data]
        while len(queue) > 0:
            data = queue.pop()
            for key, value in data.items():
                if key == target:
                    return value
                elif value == target and find_key:
                    return key
                elif type(value) == dict:
                    queue.append(value)
        return ''

    @staticmethod
    def print_table(data, title=None, no_print=None):
        if isinstance(data, requests.cookies.RequestsCookieJar):
            for cookie in iter(data):
                # FIXME ...
                pass

        if isinstance(data, dict):
            key_max_length = max([len(_) for _ in data.keys()])
            value_max_length = max([len(_) for _ in data.values()])
            title_index = (key_max_length + value_max_length) // 2
            if not title: title = 'dict to table'

            head = '-' * title_index + title.center(len(title) + 2, ' ') + '-' * title_index
            lines = [head]
            for key, value in data.items():
                formatter = "{:<%d} | {:<}" % key_max_length
                line = formatter.format(key, str(value))
                lines.append(line)

            # tail = '-' * title_index + '-' * len(title) + '-' * title_index + '--'
            # lines.append(tail)
            if no_print:
                return [_ + '\n' for _ in lines]
            else:
                for line in lines:
                    print(line)


def replacer(*rules, data=None, replace_key=False, mode='replace', dtype=None, count=False):
    result = Replacer(*rules, data=data, replace_key=replace_key, mode=mode, dtype=dtype)
    return result.data if mode != 'search' else result.search_result


def printer(data):
    return Printer(data)



import json
import logging
import requests
from requests.cookies import merge_cookies
from .data_factory import Printer, DataGroup

logger = logging.getLogger(__name__)


class ParamFactory(object):
    """
    Process requests params
    """

    # todo add cookie jar dict
    def __init__(self, url, body=None, header=None, cookie=None, overwrite=True, fmt=True):
        self._url = url
        self._path_dict = {}
        self._param_dict = {}
        self._body_dict = {}
        self._header_dict = {}
        self._cookie_dict = {}
        self.overwrite = overwrite
        self._cookie_jar = requests.cookies.RequestsCookieJar()
        self.fmt = fmt
        self.others = None

        self._path_str = self._url.split('?')[0]
        self._param_str = '' if body or '?' not in self._url else self._url.split('?')[-1]
        self._body_str = body or ''
        self._header_str = header or 'User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Safari/537.36'
        self._cookie_str = cookie or ''

        self.method = 'POST' if body else 'GET'
        if self.method == 'POST' and '=' in self._body_str:
            self.post_type = 'form'
        elif self.method == 'POST' and ':' in self._body_str:
            self.post_type = 'payload'
        else:
            self.post_type = ''

        self._path_dict = self.str_to_dict(self._url, tag='path') or {}
        self._param_dict = self.str_to_dict(self._url, tag='param') or {}
        self._body_dict = self.str_to_dict(self._body_str, tag='body') or {}
        self._header_dict = self.str_to_dict(self._header_str, tag='header') or {}
        self._cookie_dict = self.str_to_dict(self._cookie_str, tag='cookie') or {}

    def clear(self, key=None):
        if key == 'path':
            self._path_dict.clear()
        elif key == 'param':
            self._param_dict.clear()
        elif key == 'header':
            self._header_dict.clear()
        elif key == 'cookie':
            self._cookie_dict.clear()
        elif key == 'cookie_jar' and self._cookie_jar:
            self._cookie_jar.clear()
        elif key is None:
            self._path_dict.clear()
            self._param_dict.clear()
            self._header_dict.clear()
            self._cookie_dict.clear()
            self._cookie_jar.clear()

    # --------------------------------------------------------------------------------- property
    @property
    def url(self):
        if self.fmt:
            path = self.path[:-1] if self.path.endswith('/') else self.path
            return path + self.param if self.method == 'GET' else self._url
        else:
            return {'path': self.path, 'param': self.param} if self.method == 'GET' else self.path

    @property
    def path(self):
        protocol = self._path_dict.get('protocol')
        domain = self._path_dict.get('domain')
        sub_path = '/'.join([value for key, value in self._path_dict.items() if key not in ['protocol', 'domain']])
        return f'{protocol}://{domain}/{sub_path}' if self.fmt else self._path_dict

    @property
    def param(self):
        split_char = '?' if self._param_dict else ''
        param_str = '&'.join([f'{key}={value}' for key, value in self._param_dict.items()]) if self._param_dict else ''
        return split_char + param_str if self.fmt else self._param_dict

    @property
    def body(self):
        if not self.fmt:
            return self._body_dict
        elif self.post_type == 'form':
            return '&'.join([f'{key}={value}' for key, value in self._body_dict.items()])
        elif self.post_type == 'payload':
            return json.dumps(self._body_dict)
        else:
            return ''

    @property
    def headers(self):
        return self._header_dict if self._header_dict else {}

    @property
    def cookies(self):
        self._cookie_dict = self._cookie_jar.get_dict()
        return self._cookie_dict

    @property
    def cookie_jar(self):
        return self._cookie_jar

    @property
    def key_dict(self):
        return {
            'path': list(self._path_dict.keys()),
            'param': list(self._param_dict.keys()),
            'body': list(self._body_dict.keys()),
            'header': list(self._header_dict.keys()),
            'cookie': list(self._cookie_dict.keys()),
        }

    # --------------------------------------------------------------------------------- setter
    @url.setter
    def url(self, url):
        self._path_dict = self.str_to_dict(url, tag='path')
        self._param_dict = self.str_to_dict(url, tag='param')

        if self.headers.get('Host'): self.update('Host', self._path_dict.get('domain'), tag='header')

    @path.setter
    def path(self, path):
        if isinstance(path, dict):
            self._path_dict = path
        elif isinstance(path, str):
            self._path_dict = self.str_to_dict(path, tag='path')
        else:
            raise Exception('Type error ... type of path is not dict or str')

    @param.setter
    def param(self, params):
        if isinstance(params, dict):
            self._param_dict = params
        elif isinstance(params, str):
            self._param_dict = self.str_to_dict(params, tag='param')
        else:
            raise Exception('Type error ... type of params is not dict')

    @body.setter
    def body(self, body):
        if isinstance(body, dict):
            self._body_dict = body
        else:
            raise Exception('Type error ... type of body is not dict')

    @headers.setter
    def headers(self, header):
        if isinstance(header, dict):
            self._header_dict = header
        else:
            raise Exception('Type error ... type of header is not dict')

    @cookies.setter
    def cookies(self, cookie):
        if isinstance(cookie, dict):
            self._cookie_dict = cookie
        else:
            raise Exception('Type error ... type of cookie is not dict')

    # --------------------------------------------------------------------------------- core functions
    def update(self, *args, tag=None):
        """ update tag use args

        tag: param, body, header, cookie
        args: (dict,) or (str,str) or (list,list) or (dict,dict) or (list,list,dict)
        """
        # auto set tag if tag param is not be provide

        if len(args) == 1 and isinstance(args[0], dict):
            for key, value in args[0].items():
                self._update(key, value, tag=tag)

        elif len(args) == 1 and isinstance(args[0], requests.models.Response):
            self._cookie_jar.update(args[0].cookies)

        elif len(args) == 2 and isinstance(args[0], str):
            key, value = args
            self._update(key, value, tag=tag)

        elif len(args) == 2 and isinstance(args[0], list):
            for key, value in zip(args):
                self._update(key, value, tag=tag)

        elif len(args) == 2 and isinstance(args[0], dict):
            for key, value in args[0].items():
                value = args[1].get(value)
                self._update(key, value, tag=tag)
        elif len(args) == 3 and isinstance(args[0], list) and isinstance(args[2], dict):
            for key, value in zip(args[0], args[1]):
                value = args[2].get(value)
                self._update(key, value, tag=tag)

    def _update(self, key, value, tag=None):
        if not tag:
            for tag_name, key_list in self.key_dict.items():
                if key in key_list:
                    tag = tag_name
                    break
        if not tag:
            tag = 'param' if self.method == 'GET' else 'body'

        if tag == 'param':
            self._param_dict[key] = value
        elif tag == 'body':
            self._body_dict[key] = value
        elif tag == 'header':
            self._header_dict[key] = value
        elif tag == 'cookie':
            self._cookie_jar.set(key, value)
        elif tag == 'path':
            self._path_dict[key] = value

    def str_to_dict(self, string, tag=None):
        """ translate string to dict

        string: url, body, header or cookie
        tag: param, body, header or cookie
        rtype: dict
        """

        if not string: return None
        if tag == 'path' and self._path_dict: self._path_dict.clear()
        if tag == 'param' and self._param_dict: self._param_dict.clear()
        if tag == 'header' and self._header_dict: self._header_dict.clear()
        if tag == 'cookie' and self._cookie_dict: self._cookie_dict.clear()

        string = string.strip()

        if tag == 'path' and '://' in string:
            if '?' in string: string = string.split('?')[0]

            protocol = string.split('://', 1)[0]
            domain = string.split('://', 1)[1].split('/')[0]
            path_list = string.split('://', 1)[1].split('/')[1:]

            self._path_dict['protocol'] = protocol
            self._path_dict['domain'] = domain
            self._path_dict = {**self._path_dict, **dict(zip([str(i + 1) for i in range(len(path_list))], path_list))}
            return self._path_dict

        if tag == 'param':
            if self.body: return

            # sting : 'https://...?...'
            if '?' in string:
                if '=' in string:
                    string = string.split('?')[-1]
                    self._param_dict = dict(
                        [_.split('=', 1) for _ in string.split('&') if '=' in _ and not _.endswith('=')])
                    self._param_dict = {**self._param_dict, **dict([(_.strip('='), '') for _ in string.split('&')
                                                                    if _.endswith('=') or '=' not in _])}
                else:
                    self._param_dict = dict([(string, string) for _ in string.split('&')])

            # sting : 'name=...'
            elif '=' in string:
                self._param_dict = dict([_.split('=', 1) for _ in string.split('&')])
            else:
                self._param_dict = {}

            return self._param_dict

        if tag == 'body':
            if '=' in string and '&' in string:
                self._body_dict = dict([_.split('=', 1) for _ in string.split('&')])
            elif '=' in string:
                self._body_dict = dict([string.split('=')])
            elif ':' in string:
                self._body_dict = json.loads(string)
            return self._body_dict

        if tag == 'header' or tag == 'cookie':
            split_params = ['\n', ':'] if tag == 'header' else [';', '=']
            target_dict = {}
            for field in string.split(split_params[0]):
                keys = target_dict.keys()

                key, value = field.split(split_params[1], 1)
                key, value = [key.strip(), value.strip()]

                value_in_dict = target_dict.get(key)
                if key in keys and isinstance(value_in_dict, str):
                    target_dict[key] = [value_in_dict, value]
                elif isinstance(value_in_dict, list):
                    target_dict[key].append(value)
                else:
                    target_dict[key] = value

            if tag == 'header': self._header_dict = target_dict
            if tag == 'cookie': self._cookie_dict = target_dict
            return target_dict

    def preview(self, tag=None):
        d_group = DataGroup(name='Params')
        d_group.add_info('url', self.url)
        d_group.add_info('param', self.param)
        d_group.add_info('body', self.body)

        preview_dict = {'headers': self.headers, 'cookies': self.cookies}
        preview_list = ['headers', 'cookies'] if not tag else [tag]

        for key, value in preview_dict.items():
            if key in preview_list:
                d_group.add_data(key, value)
            else:
                continue

        printer = Printer()
        return printer.parse_data_group(d_group)

    def __repr__(self):
        lines = self.preview()
        return '\n'.join(lines)
